#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          02-07-2018
//      Created for Project :   FLAME
"""


from pyramid.httpexceptions import HTTPBadRequest, HTTPConflict, HTTPNotFound
from pyramid.view import view_defaults, view_config
from clmcservice.models import ServiceFunctionChain
from clmcservice.configapi.utilities import validate_sfchain_body, validate_sfchain_params


@view_defaults(renderer='json')
class SFCConfigAPI(object):
    """
    A class-based view for posting and retrieving configuration data for service function chains to the CLMC service
    """

    def __init__(self, request):
        """
        Initialises the instance of the view with the request argument.

        :param request: client's call request
        """

        self.request = request

    @view_config(route_name='config_sfc', request_method='GET')
    def get_all(self):
        """
        GET API call for all resources.

        :return: A list of all service function chains found in the database.
        """

        return [instance.json for instance in ServiceFunctionChain.query()]

    @view_config(route_name='config_sfc_instance', request_method='GET')
    def get_one(self):
        """
        GET API call for a single resources.

        :return: One service function chain instance retrieved from the database by querying the uniquely constrained columns.
        :raises HTTPBadRequest: if the request parameters are invalid(invalid url query string)
        :raises HTTPNotFound: if a resource with the given parameters doesn't exist in the database
        """

        sf_chain = self._get_sf_chain_from_url_string()
        if sf_chain is None:
            raise HTTPNotFound("A service function chain with the given parameters doesn't exist.")
        else:
            return sf_chain.json

    @view_config(route_name='config_sfc', request_method='POST')
    def post(self):
        """
        A POST API call to create a new service function chain.

        :return: A JSON response to the POST call - essentially with the data of the new resource
        :raises HTTPBadRequest: if request body is not a valid JSON for the service function chain
        :raises HTTPConflict: if the unique constraints are not preserved after the creation of a new instance
        """

        # create an instance of the model and add it to the database table
        sf_chain = self._validate_and_create()
        json_data = sf_chain.json
        ServiceFunctionChain.add(sf_chain)

        self.request.response.status = 201

        return json_data

    @view_config(route_name='config_sfc_instance', request_method='PUT')
    def put(self):
        """
        A PUT API call to update a service function chain.

        :return: A JSON response representing the updated object
        :raises HTTPBadRequest: if the request parameters are invalid(invalid url query string)
        :raises HTTPNotFound: if a resource with the given parameters doesn't exist in the database
        """

        sf_chain = self._get_sf_chain_from_url_string()
        if sf_chain is None:
            raise HTTPNotFound("A service function chain with the given parameters doesn't exist.")
        else:
            try:
                body = self.request.body.decode(self.request.charset)
                validated_body = validate_sfchain_body(body)  # validate the content and receive a json dictionary object
            except AssertionError as e:
                raise HTTPBadRequest("Bad request content. Service function chain format is incorrect: {0}".format(e.args))

            new_resource = validated_body
            old_resource = sf_chain.json
            updating = new_resource["sfc"] == old_resource["sfc"] and new_resource["sfc_i"] == old_resource["sfc_i"]

            if updating:
                ServiceFunctionChain.delete(sf_chain)
                new_sf_chain = ServiceFunctionChain(**validated_body)
                ServiceFunctionChain.add(new_sf_chain)
            else:
                resource_exists = ServiceFunctionChain.exists(new_resource["sfc"], new_resource["sfc_i"])
                if resource_exists:
                    raise HTTPConflict("Service function chain with this data already exists.")  # error 409 in case of resource conflict

                new_sf_chain = ServiceFunctionChain(**validated_body)
                ServiceFunctionChain.replace(sf_chain, new_sf_chain)

            return validated_body

    @view_config(route_name='config_sfc_instance', request_method='DELETE')
    def delete(self):
        """
        Deletes an instance of a service function chain in the database.

        :return: An content of the object that has been deleted
        :raises HTTPBadRequest: if the request parameters are invalid(invalid url query string)
        :raises HTTPNotFound: if a resource with the given parameters doesn't exist in the database
        """

        sf_chain = self._get_sf_chain_from_url_string()
        if sf_chain is None:
            raise HTTPNotFound("A service function chain with the given parameters doesn't exist.")
        else:
            deleted = sf_chain.json
            ServiceFunctionChain.delete(sf_chain)
            return deleted

    def _get_sf_chain_from_url_string(self):
        """
        Retrieves a service function chain from the database by validating and then using the request url parameters.

        :return: An instance of a service function chain or None if not existing
        """

        params = {}
        for attribute in ServiceFunctionChain.constrained_columns():
            if attribute in self.request.params:
                params[attribute] = self.request.params.get(attribute)

        try:
            params = validate_sfchain_params(params)
        except AssertionError as e:
            raise HTTPBadRequest("Request format is incorrect: {0}".format(e.args))

        sf_chain = ServiceFunctionChain.get(**params)
        return sf_chain

    def _validate_and_create(self):
        """
        Validates the request body and checks if a resource with the given attributes already exists.

        :return: a new instance of the model, if the resource doesn't exist
        :raises HTTPBadRequest: if request body is not a valid JSON for the service function chain
        :raises HTTPConflict: if the unique constraints are not preserved after the creation of a new instance
        """

        try:
            body = self.request.body.decode(self.request.charset)
            validated_body = validate_sfchain_body(body)  # validate the content and receive a json dictionary object
        except AssertionError as e:
            raise HTTPBadRequest("Bad request content. Service function chain format is incorrect: {0}".format(e.args))

        resource = validated_body

        resource_exists = ServiceFunctionChain.exists(resource["sfc"], resource["sfc_i"])
        if resource_exists:
            raise HTTPConflict("Service function chain with this data already exists.")  # error 409 in case of resource conflict

        # create an instance of the model
        sf_chain = ServiceFunctionChain(**resource)

        return sf_chain
